package webapp

import (
	"net/http"
	"net/http/httptest"
	"os"
	"testing"
	"strings"

	"github.com/DATA-DOG/go-sqlmock"
	"gontrol/src/logger"
)

func newTestApp(t *testing.T) (*App, sqlmock.Sqlmock) {
	t.Helper()

	tmpl, err := ParseTemplates()
	if err != nil {
		t.Fatalf("ParseTemplates: %v", err)
	}

	db, mock, err := sqlmock.New()
	if err != nil {
		t.Fatalf("sqlmock: %v", err)
	}
	return &App{Tmpl: tmpl, DB: db}, mock
}

// initFakeLogs puts one record in the in‑memory log DB used by logger.FetchLogs.
func initFakeLogs(t *testing.T) {
	t.Helper()
	if err := logger.InsertLog(logger.Info, "unit‑test"); err != nil {
		t.Fatalf("InsertLog: %v", err)
	}
}

func TestLogsHandler_HTML(t *testing.T) {
	tmp, err := os.CreateTemp("", "logs_*.db")
	if err != nil {
		t.Fatalf("Temp DB: %v", err)
	}
	tmp.Close()
	defer os.Remove(tmp.Name())

	if err := logger.InitDB(tmp.Name()); err != nil {
		t.Fatalf("logger.InitDB: %v", err)
	}
	defer logger.CloseDB()

	logger.InsertLog(logger.Error, "fake-log-1")

	app, mock := newTestApp(t)
	defer app.DB.Close()

	if err := mock.ExpectationsWereMet(); err != nil {
		t.Fatalf("sqlmock expectations: %v", err)
	}


	req := httptest.NewRequest(http.MethodGet, "/logs/error?limit=1", nil)
	rec := httptest.NewRecorder()

	app.logsHandler(rec, req)

	if rec.Code != http.StatusOK {
		t.Fatalf("status = %d; want 200", rec.Code)
	}
	if ct := rec.Header().Get("Content-Type"); ct != "text/html; charset=utf-8" {
		t.Errorf("Content-Type = %s; want text/html", ct)
	}
	if !strings.Contains(rec.Body.String(), "fake-log-1") {
		t.Errorf("rendered HTML missing our fake log line")
	}
}

// Test that `limit=abc` is rejected with 400 and FetchLogs is **not** hit.
func TestLogsHandler_InvalidLimit(t *testing.T) {
	initFakeLogs(t)

	app, _ := newTestApp(t) // helper from previous file

	req := httptest.NewRequest(http.MethodGet, "/logs?limit=abc", nil)
	rec := httptest.NewRecorder()

	app.logsHandler(rec, req)

	if rec.Code != http.StatusBadRequest {
		t.Fatalf("status = %d; want 400", rec.Code)
	}
	if !strings.Contains(rec.Body.String(), "Invalid count value") {
		t.Errorf("missing error message")
	}
}

// Test JSON response when client sends Accept: application/json.
func TestLogsHandler_JSON(t *testing.T) {
	initFakeLogs(t)

	app, _ := newTestApp(t)

	req := httptest.NewRequest(http.MethodGet, "/logs?limit=1", nil)
	req.Header.Set("Accept", "application/json")
	rec := httptest.NewRecorder()

	app.logsHandler(rec, req)

	if rec.Code != http.StatusOK {
		t.Fatalf("status = %d; want 200", rec.Code)
	}
	if ct := rec.Header().Get("Content-Type"); ct != "application/json" {
		t.Errorf("Content‑Type = %q; want application/json", ct)
	}
	if !strings.Contains(rec.Body.String(), "unit‑test") {
		t.Errorf("JSON body missing expected log entry")
	}
}
